/*
 * liba52_adapter.c
 * Copyright (C) 2010 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
//#define __DEBUG__

#include <windows.h>

#include <libavutil/internal.h>
#include "ffmpeg_stub.h"

#ifdef fprintf
#undef fprintf
#endif

#ifdef restrict
#undef restrict
#endif

#include <a52dec/config.h>
#include "liba52_stub.h"
#include <a52dec/mm_accel.h>
#include <a52dec/audio_out.h>
#include "audio_out_internal.h"

#ifndef __CS__
#define InitializeCriticalSection(cs)
#define DeleteCriticalSection(cs)
#define EnterCriticalSection(cs)
#define LeaveCriticalSection(cs)
#else // __CS__
static CRITICAL_SECTION cs;
#endif // __CS__

typedef struct {
  int disable_accel;
  int disable_dynrng;
  int disable_adjust;
  sample_t gain;

  int sample_rate;
  int flags;

  uint8_t buf[3840];
  uint8_t *bufptr;
  uint8_t *bufpos;

  a52_state_t *state;
  ao_instance_t *output;
} LibA52AdapterContext;

static av_cold int liba52_adapter_init(AVCodecContext *avctx)
{
  int ret=-1;
  LibA52AdapterContext *priv=avctx->priv_data;
	
#ifdef __DEBUG__
  fprintf(stderr, "liba52_adapter_init(%p): %d, %d\n", avctx,
      avctx->sample_rate, avctx->channels);
#endif

#ifndef OLD
  avctx->sample_fmt=AV_SAMPLE_FMT_S32;
#else
  avctx->sample_fmt=SAMPLE_FMT_S32;
#endif
  //avctx->sample_rate=48000;
  avctx->channels=2;

  priv->disable_accel=0;
  //priv->disable_accel=1;
  priv->disable_dynrng=1;
  priv->disable_adjust=0;
  priv->gain=1;
  priv->state=NULL;
  priv->output=NULL;

  priv->sample_rate=0;
  priv->flags=0;

  priv->bufptr=priv->buf;
  priv->bufpos=priv->buf + 7;

  EnterCriticalSection(&cs);

  if (NULL==(priv->state=__a52_init(priv->disable_accel?0:MM_ACCEL_DJBFFT)))
    goto cleanup;

  if (NULL==(priv->output=ao_raw32_open()))
    goto cleanup;

  ret=0;
cleanup:
  LeaveCriticalSection(&cs);

  return ret;
}

void liba52_adapter_set_gain(AVCodecContext *avctx, double gain)
{
  ((LibA52AdapterContext *)avctx->priv_data)->gain=gain;
}

static av_cold int liba52_adapter_close(AVCodecContext *avctx)
{
  LibA52AdapterContext *priv=avctx->priv_data;

#ifdef __DEBUG__
  fprintf(stderr, "liba52_adapter_close()\n");
#endif

  EnterCriticalSection(&cs);

  if (NULL!=priv->output)
    ao_close(priv->output);

  if (NULL!=priv->state)
    __a52_free(priv->state);

  LeaveCriticalSection(&cs);

  return 0;
}

static int liba52_adapter_decode(AVCodecContext *avctx, void *data,
		int *data_size, AVPacket *avpkt)
{
  LibA52AdapterContext *priv=avctx->priv_data;

  uint8_t *start=avpkt->data;
  uint8_t *end=start+avpkt->size;

  int bit_rate;
  int len;
  int ret;

#ifdef __DEBUG__
  //fprintf(stderr, "liba52_adapter_decode(): %d\n", avpkt->size);
#endif
  ///////////////////
  ((raw32_instance_t *)priv->output)->int32_samples=(int32_t *)data;
  ((raw32_instance_t *)priv->output)->offs=0;
  ((raw32_instance_t *)priv->output)->size=0;
  ((raw32_instance_t *)priv->output)->max_size=*data_size;
  ///////////////////

  while (1) {
    len = end - start;
    if (!len)
      break;
    if (len > priv->bufpos - priv->bufptr)
      len = priv->bufpos - priv->bufptr;
#ifdef __DEBUG__
  //fprintf(stderr, "+++ %d: %d\n", BUF_SIZE, (priv->bufptr+len)-priv->buf);
#endif
    memcpy (priv->bufptr, start, len);
    priv->bufptr += len;
    start += len;
    if (priv->bufptr == priv->bufpos) {
      if (priv->bufpos == priv->buf + 7) {
        int length;

        EnterCriticalSection(&cs);
        length = __a52_syncinfo (priv->buf, &priv->flags, &priv->sample_rate,
		    &bit_rate);
        LeaveCriticalSection(&cs);
        if (!length) {
#ifdef __DEBUG__
          fprintf (stderr, "skip\n");
#endif
          for (priv->bufptr = priv->buf; priv->bufptr < priv->buf + 6;
		      priv->bufptr++)
            priv->bufptr[0] = priv->bufptr[1];
          continue;
        }
        priv->bufpos = priv->buf + length;
        ///////////////////
		avctx->sample_rate=priv->sample_rate;
        ///////////////////
      } else {
        sample_t level, bias;
        int i;

        if (ao_setup (priv->output, priv->sample_rate, &priv->flags,
		    &level, &bias))
          goto error;
        if (!priv->disable_adjust)
          priv->flags |= A52_ADJUST_LEVEL;
        level *= priv->gain;
        EnterCriticalSection(&cs);
        ret=__a52_frame (priv->state, priv->buf, &priv->flags, &level, bias);
        LeaveCriticalSection(&cs);
        if (ret)
          goto error;
        if (priv->disable_dynrng) {
          EnterCriticalSection(&cs);
          __a52_dynrng (priv->state, NULL, NULL);
          LeaveCriticalSection(&cs);
		}

        for (i = 0; i < 6; i++) {
          EnterCriticalSection(&cs);
          ret=__a52_block (priv->state);
          LeaveCriticalSection(&cs);
          if (ret)
            goto error;
          EnterCriticalSection(&cs);
          ret=ao_play (priv->output, priv->flags, __a52_samples (priv->state));
          LeaveCriticalSection(&cs);
          if (ret)
            goto error;
        }
        priv->bufptr = priv->buf;
        priv->bufpos = priv->buf + 7;
#ifdef __XXX__
        print_fps (0);
#endif
        continue;
      error:
#ifdef __DEBUG__
        fprintf (stderr, "error\n");
#endif
        priv->bufptr = priv->buf;
        priv->bufpos = priv->buf + 7;
      }
    }
  }

  ///////////////////
  *data_size=((raw32_instance_t *)priv->output)->size;
  ///////////////////

  return avpkt->size;
}

void liba52_adapter_start()
{
  InitializeCriticalSection(&cs);
}

void liba52_adapter_stop()
{
  DeleteCriticalSection(&cs);
}

AVCodec liba52_adapter = {
  .name="liba52",
  .type=AVMEDIA_TYPE_AUDIO,
  .id=CODEC_ID_AC3,
  .priv_data_size=sizeof(LibA52AdapterContext),
  .init=liba52_adapter_init,
  .close=liba52_adapter_close,
  .decode=liba52_adapter_decode,
  .long_name=NULL_IF_CONFIG_SMALL("ATSC A/52A (AC-3)"),
};
