/*
 * libmad_adapter.c
 * Copyright (C) 2010 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <libavutil/internal.h>
#include "ffmpeg_stub.h"

# include "madutil.h"
# include "libmad_stub.h"
# include "ffmpeg_stub.h"

# ifdef fprintf
# undef fprintf
# endif

# define MPEG_BUFSZ	40000	/* 2.5 s at 128 kbps; 1 s at 320 kbps */
# define MIN(a,b) ((a)<(b)?(a):(b))

typedef struct {
  struct mad_decoder decoder;
  mad_sync_t sync;
  double gain;

  struct input {
    AVPacket *avpkt;

    unsigned char *data;
    unsigned long length;

    int eof;
  } input;

  struct output {
    unsigned int bitrate, samplerate, nchannels, nsamples;
    int32_t *data;
	int data_size;
  } output;
} LibMadAdapterContext;

/*
 * NAME:	decode->input_read()
 * DESCRIPTION:	(re)fill decoder input buffer by reading a file descriptor
 */
static
enum mad_flow decode_input_read(void *data, struct mad_stream *stream)
{
  LibMadAdapterContext *ctx=data;
  struct input *input = &ctx->input;
  int len;

  if (input->eof)
    return MAD_FLOW_STOP;

  if (stream->next_frame) {
    memmove(input->data, stream->next_frame,
	    input->length = &input->data[input->length] - stream->next_frame);
  }

  len=MIN(MPEG_BUFSZ-input->length-MAD_BUFFER_GUARD,input->avpkt->size);
  memcpy(input->data + input->length, input->avpkt->data, len);

  if (len == 0) {
    input->eof = 1;

    assert(MPEG_BUFSZ - input->length >= MAD_BUFFER_GUARD);

    while (len < MAD_BUFFER_GUARD)
      input->data[input->length + len++] = 0;
  }

  __mad_stream_buffer(stream, input->data, input->length += len);

  return MAD_FLOW_CONTINUE;
}

static inline
signed int scale(mad_fixed_t sample)
{
  /* clip */
  if (sample >= MAD_F_ONE)
    sample = MAD_F_ONE - 1;
  else if (sample < -MAD_F_ONE)
    sample = -MAD_F_ONE;

  return sample << 3;
}

/*
 * NAME:	decode->output()
 * DESCRIPTION: configure audio module and output decoded samples
 */
static
enum mad_flow decode_output(void *data, struct mad_header const *header,
			    struct mad_pcm *pcm)
{
  LibMadAdapterContext *ctx=data;
  struct output *output=&ctx->output;
  unsigned int nchannels, nsamples;
  mad_fixed_t const *left_ch, *right_ch;

  /* pcm->samplerate contains the sampling frequency */
  nchannels = pcm->channels;
  nsamples  = pcm->length;
  left_ch   = pcm->samples[0];
  right_ch  = pcm->samples[1];

  output->bitrate=header->bitrate;
  output->samplerate=pcm->samplerate;
  output->nchannels=nchannels;
  output->nsamples+=nsamples;

  if (1.0==ctx->gain) {
    while (nsamples--) {
	  *output->data++=scale(*left_ch++);
      output->data_size+=sizeof *output->data;

      if (nchannels == 2) {
        *output->data++=scale(*right_ch++);
        output->data_size+=sizeof *output->data;
	  }
    }
  }
  else {
	mad_fixed_t gain=mad_f_tofixed(ctx->gain);

    while (nsamples--) {
	  *output->data++=scale(mad_f_mul(gain,*left_ch++));
      output->data_size+=sizeof *output->data;

      if (nchannels == 2) {
        *output->data++=scale(mad_f_mul(gain,*right_ch++));
        output->data_size+=sizeof *output->data;
	  }
    }
  }

  return MAD_FLOW_CONTINUE;
}

void libmad_adapter_set_gain(AVCodecContext *avctx, double gain)
{
  ((LibMadAdapterContext *)avctx->priv_data)->gain=gain;
}

static av_cold int decode_init(AVCodecContext * avctx)
{
  LibMadAdapterContext *ctx=avctx->priv_data;

  __mad_decoder_init(&ctx->decoder, ctx,
		   decode_input_read,
		   0, 0,
		   decode_output,
		   0, 0);
  ctx->gain=1.0;
  ctx->input.data=__av_malloc(MPEG_BUFSZ);
  ctx->input.length=0;
  ctx->decoder.sync=__av_malloc(sizeof(*ctx->decoder.sync));
  mad_sync_init(&ctx->sync, &ctx->decoder);
  return 0;
}

static av_cold int decode_close(AVCodecContext *avctx)
{
  LibMadAdapterContext *ctx=avctx->priv_data;
  mad_sync_finish(&ctx->sync);
  __av_free(ctx->decoder.sync);
  __av_free(ctx->input.data);
  ctx->decoder.sync=NULL;
  __mad_decoder_finish(&ctx->decoder);
  return -1;
}

static int decode_frame(AVCodecContext * avctx,
                        void *data, int *data_size,
                        AVPacket *avpkt)
{
  LibMadAdapterContext *ctx=avctx->priv_data;
  mad_sync_t *sync=&ctx->sync;

  if (MAD_SYNC_STATE_OUTER_LOOP!=sync->state)
    return -1;

  ctx->input.avpkt=avpkt;
  ctx->output.data=data;
  ctx->output.data_size=0;

  do {
    mad_sync_next(sync);
  } while (MAD_SYNC_STATE_INNER_LOOP==sync->state);

  avctx->bit_rate=ctx->output.bitrate;
#ifndef OLD
  avctx->sample_fmt=AV_SAMPLE_FMT_S32;
#else
  avctx->sample_fmt=SAMPLE_FMT_S32;
#endif
  avctx->sample_rate=ctx->output.samplerate;
  avctx->channels=ctx->output.nchannels;

  *data_size=ctx->output.data_size;

  if (ctx->input.eof&&MAD_SYNC_STATE_OUTER_LOOP==sync->state)
    mad_sync_next(sync);

  return avpkt->size;
}

AVCodec libmad_adapter =
{
  .name=LIBMAD,
  .type=AVMEDIA_TYPE_AUDIO,
  .id=CODEC_ID_MP3,
  .priv_data_size=sizeof(LibMadAdapterContext),
  .init=decode_init,
  .close=decode_close,
  .decode=decode_frame,
  .long_name=NULL_IF_CONFIG_SMALL("MP3 (MPEG audio layer 3)")
};
