/*
 * mad_sync.c
 * Copyright (C) 2010 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
# include <assert.h>
# include "madutil.h"
# include "libmad_stub.h"

static
enum mad_flow error_default(void *data, struct mad_stream *stream,
			    struct mad_frame *frame)
{
  int *bad_last_frame = data;

  switch (stream->error) {
  case MAD_ERROR_BADCRC:
    if (*bad_last_frame)
      __mad_frame_mute(frame);
    else
      *bad_last_frame = 1;

    return MAD_FLOW_IGNORE;

  default:
    return MAD_FLOW_CONTINUE;
  }
}

int mad_sync_init(mad_sync_t *sync, struct mad_decoder *decoder)
{
  sync->state=MAD_SYNC_STATE_UNDEFINED;
  sync->decoder=decoder;

  sync->bad_last_frame = 0;

  if (decoder->input_func == 0)
    return 0;

  if (decoder->error_func) {
    sync->error_func = decoder->error_func;
    sync->error_data = decoder->cb_data;
  }
  else {
    sync->error_func = error_default;
    sync->error_data = &sync->bad_last_frame;
  }

  sync->stream = &decoder->sync->stream;
  sync->frame  = &decoder->sync->frame;
  sync->synth  = &decoder->sync->synth;

  __mad_stream_init(sync->stream);
  __mad_frame_init(sync->frame);
  __mad_synth_init(sync->synth);

  mad_stream_options(sync->stream, decoder->options);
  sync->state=MAD_SYNC_STATE_OUTER_LOOP;

  return 1;
}

void mad_sync_finish(mad_sync_t *sync)
{
  mad_synth_finish(sync->synth);
  __mad_frame_finish(sync->frame);
  __mad_stream_finish(sync->stream);
}

static
void mad_sync_outer_loop(mad_sync_t *sync)
{
  struct mad_decoder *decoder=sync->decoder;

  assert(MAD_SYNC_STATE_OUTER_LOOP==sync->state);

  switch (decoder->input_func(decoder->cb_data, sync->stream)) {
  case MAD_FLOW_STOP:
    sync->state=MAD_SYNC_STATE_DONE;
    return;
  case MAD_FLOW_BREAK:
    sync->state=MAD_SYNC_STATE_FAIL;
    return;
  case MAD_FLOW_IGNORE:
    return;
  case MAD_FLOW_CONTINUE:
    sync->state=MAD_SYNC_STATE_INNER_LOOP;
    break;
  }
}

# define MAD_SYNC_MACHINE_INNER_LOOP_BREAK(error) \
  ((error)==MAD_ERROR_BUFLEN?MAD_SYNC_STATE_OUTER_LOOP:MAD_SYNC_STATE_FAIL)

static
void mad_sync_inner_loop(mad_sync_t *sync)
{
  struct mad_decoder *decoder=sync->decoder;
  void *error_data=sync->error_data;
  struct mad_stream *stream=sync->stream;
  struct mad_frame *frame=sync->frame;
  struct mad_synth *synth=sync->synth;

  assert(MAD_SYNC_STATE_INNER_LOOP==sync->state);

# if defined(USE_ASYNC)
  if (decoder->mode == MAD_DECODER_MODE_ASYNC) {
    switch (check_message(decoder)) {
    case MAD_FLOW_IGNORE:
    case MAD_FLOW_CONTINUE:
      break;
    case MAD_FLOW_BREAK:
      sync->state=MAD_SYNC_STATE_FAIL;
      return;
    case MAD_FLOW_STOP:
      sync->state=MAD_SYNC_STATE_DONE;
      return;
    }
  }
# endif

  if (decoder->header_func) {
    if (__mad_header_decode(&frame->header, stream) == -1) {
      if (!MAD_RECOVERABLE(stream->error)) {
        sync->state=MAD_SYNC_MACHINE_INNER_LOOP_BREAK(stream->error);
        return;
      }

      switch (sync->error_func(error_data, stream, frame)) {
      case MAD_FLOW_STOP:
        sync->state=MAD_SYNC_STATE_DONE;
        return;
      case MAD_FLOW_BREAK:
        sync->state=MAD_SYNC_STATE_FAIL;
        return;
      case MAD_FLOW_IGNORE:
      case MAD_FLOW_CONTINUE:
      default:
        return;
      }
    }

    switch (decoder->header_func(decoder->cb_data, &frame->header)) {
    case MAD_FLOW_STOP:
      sync->state=MAD_SYNC_STATE_DONE;
      return;
    case MAD_FLOW_BREAK:
      sync->state=MAD_SYNC_STATE_FAIL;
      return;
    case MAD_FLOW_IGNORE:
      return;
    case MAD_FLOW_CONTINUE:
      break;
    }
  }

  if (__mad_frame_decode(frame, stream) == -1) {
    if (!MAD_RECOVERABLE(stream->error)) {
      sync->state=MAD_SYNC_MACHINE_INNER_LOOP_BREAK(stream->error);
      return;
    }

    switch (sync->error_func(error_data, stream, frame)) {
    case MAD_FLOW_STOP:
      sync->state=MAD_SYNC_STATE_DONE;
      return;
    case MAD_FLOW_BREAK:
      sync->state=MAD_SYNC_STATE_FAIL;
      return;
    case MAD_FLOW_IGNORE:
      break;
    case MAD_FLOW_CONTINUE:
    default:
      return;
    }
  }
  else
    sync->bad_last_frame = 0;

  if (decoder->filter_func) {
    switch (decoder->filter_func(decoder->cb_data, stream, frame)) {
    case MAD_FLOW_STOP:
      sync->state=MAD_SYNC_STATE_DONE;
      return;
    case MAD_FLOW_BREAK:
      sync->state=MAD_SYNC_STATE_FAIL;
      return;
    case MAD_FLOW_IGNORE:
      return;
    case MAD_FLOW_CONTINUE:
      break;
    }
  }

  __mad_synth_frame(synth, frame);

  if (decoder->output_func) {
    switch (decoder->output_func(decoder->cb_data,
                                 &frame->header, &synth->pcm)) {
    case MAD_FLOW_STOP:
      sync->state=MAD_SYNC_STATE_DONE;
      return;
    case MAD_FLOW_BREAK:
      sync->state=MAD_SYNC_STATE_FAIL;
      return;
    case MAD_FLOW_IGNORE:
    case MAD_FLOW_CONTINUE:
      break;
    }
  }
}

int mad_sync_next(mad_sync_t *sync)
{
  switch (sync->state) {
  case MAD_SYNC_STATE_OUTER_LOOP:
    mad_sync_outer_loop(sync);
    return 1;
  case MAD_SYNC_STATE_INNER_LOOP:
    mad_sync_inner_loop(sync);
    return 1;
  case MAD_SYNC_STATE_DONE:
    return 0;
  case MAD_SYNC_STATE_FAIL:
  default:
    return 0;
  }
}
